import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter_hooks/flutter_hooks.dart';
import 'package:flutter_query/flutter_query.dart';
import 'package:flutter_query/src/query.dart';
import 'package:flutter_query/src/query_cache.dart';
import 'package:flutter_query/src/query_observer.dart';
import 'package:provider/provider.dart';

typedef Future<T> QueryFunction<T>();
typedef Future PageableQueryFunction<T>(int page);
typedef dynamic GetFetchMoreFunction(latestPage);

typedef void QueryCallback<T>(T data);

QueryResult<T> useQuery<T>(
  String key,
  QueryFunction<T> fn, {
  QueryCallback onSuccess,
  QueryCallback onError,
  QueryCallback onSettled,
}) {
  return use(
    _QueryHook<T>(
      QueryConfig(
        queryKey: key,
        fn: fn,
        onSuccess: onSuccess,
        onError: onError,
        onSettled: onSettled,
      ),
    ),
  );
}

QueryResult<T> useInfiniteQuery<T>(
  String key,
  PageableQueryFunction<T> fn, {
  GetFetchMoreFunction getFetchMore,
  QueryCallback onSuccess,
  QueryCallback onError,
  QueryCallback onSettled,
}) {
  return use(
    _QueryHook<T>(
      QueryConfig(
        queryKey: key,
        fn: fn,
        onSuccess: onSuccess,
        onError: onError,
        onSettled: onSettled,
        getFetchMore: getFetchMore,
        isInfinite: true,
      ),
    ),
  );
}

class _QueryHook<T> extends Hook<QueryResult<T>> {
  const _QueryHook(this.config);

  final QueryConfig config;

  @override
  _QueryStateHook<T> createState() => _QueryStateHook();
}

class _QueryStateHook<T> extends HookState<QueryResult<T>, _QueryHook<T>> {
  QueryResult<T> _result;
  Function _unsubscribe;

  @override
  void initHook() {
    final cache = Provider.of<QueryCache>(context, listen: false);
    _unsubscribe = cache.watchQuery(
      hook.config,
      (QueryResult<T> result) {
        setState(() {
          _result = result;
        });
      },
    );
    super.initHook();
  }

  @override
  void dispose() {
    _unsubscribe();
    super.dispose();
  }

  @override
  QueryResult<T> build(BuildContext context) {
    return _result;
  }
}
