import 'dart:async';

import 'package:flutter/material.dart';
import 'package:rxdart/subjects.dart';

class Query<T> {
  final BehaviorSubject<QueryState<T>> stateStream =
      BehaviorSubject.seeded(QueryState<T>(), sync: true);

  QueryConfig config;

  Query(this.config);

  Future<T> fetch() {
    Future<T> queryFn;
    if (!stateStream.value.hasData) {
      stateStream.add(QueryState(
        snapshot: AsyncSnapshot<T>.waiting(),
      ));
    }

    if (config.isInfinite) {
      final cursor = config.getFetchMore(stateStream.value.data);
      if (cursor != false) {
        final current = stateStream.value;
        // TODO: Find a better way to do this than just mutating state
        current.isFetchingMore = true;
        stateStream.add(current);
        queryFn = config.fn(cursor);
      }
    } else {
      queryFn = config.fn();
    }

    return queryFn.then((data) {
      var current = stateStream.value.data;
      var canFetchMore;
      if (config.isInfinite) {
        if (current is List<T> && data is List<T>) {
          data.insertAll(0, current as dynamic);
        }
        canFetchMore = data is List ? data.length > 0 : false;
      }

      stateStream.add(
        QueryState(
          snapshot: AsyncSnapshot<T>.withData(ConnectionState.active, data),
          updatedAt: DateTime.now().millisecondsSinceEpoch,
          canFetchMore: canFetchMore,
          failureCount: stateStream?.value?.failureCount ?? 0,
          isFetchingMore: false,
        ),
      );
      return data;
    }).catchError((err) {
      stateStream.add(
        QueryState(
          snapshot: AsyncSnapshot<T>.withError(ConnectionState.done, err),
          updatedAt: DateTime.now().millisecondsSinceEpoch,
          failureCount: (stateStream?.value?.failureCount ?? 0) + 1,
        ),
      );
    });
  }

  bool isStaleByTime([staleTime = 0]) {
    return (stateStream.value.isInvalidated ||
        stateStream.value.isSuccess ||
        (stateStream.value.updatedAt ?? 0) + staleTime <=
            DateTime.now().millisecondsSinceEpoch);
  }

  void invalidate() {
    stateStream.add(
      QueryState(
        snapshot: AsyncSnapshot<T>.withData(
          ConnectionState.active,
          stateStream.value?.data,
        ),
        updatedAt: DateTime.now().millisecondsSinceEpoch,
        canFetchMore: stateStream.value?.canFetchMore ?? false,
        failureCount: stateStream?.value?.failureCount ?? 0,
      ),
    );
  }

  void dispose() {
    stateStream.close();
  }
}

class QueryState<T> {
  AsyncSnapshot<T> snapshot;
  int updatedAt;
  int failureCount;
  bool isInvalidated;
  bool canFetchMore;
  bool isFetchingMore;

  bool get isError {
    return snapshot?.hasError;
  }

  bool get isFetching {
    return snapshot?.connectionState == ConnectionState.waiting;
  }

  bool get isIdle {
    // print('isIdle needs implementation');
    return false;
  }

  bool get isLoading {
    return isFetching && !(snapshot?.hasData ?? false);
  }

  bool get isSuccess {
    return snapshot?.connectionState == ConnectionState.active &&
        !isError &&
        (snapshot?.hasData ?? false);
  }
  //  isPreviousData,
  //  isStale,

  T get data {
    return snapshot.data;
  }

  bool get hasData {
    return snapshot.data != null;
  }

  QueryState({
    this.snapshot = const AsyncSnapshot.nothing(),
    this.failureCount = 0,
    this.canFetchMore = false,
    this.isFetchingMore = false,
    this.updatedAt,
  });
}

class QueryConfig {
  final Function fn;
  final Function getFetchMore;
  final bool isInfinite;
  final bool enabled;
  final String queryKey;
  final Function onSuccess;
  final Function onError;
  final Function onSettled;
  final int staleTime;

  QueryConfig({
    @required this.fn,
    this.getFetchMore,
    this.queryKey,
    this.onSuccess,
    this.onError,
    this.onSettled,
    this.isInfinite = false,
    this.enabled = true,
    this.staleTime = 0,
  });
}
