import 'package:flutter/widgets.dart';
import 'package:flutter_query/src/query_observer.dart';
import 'package:provider/provider.dart';

import '../flutter_query.dart';

class QueryCache {
  Map<String, Query> _queryMap = {};

  bool get isFetching {
    return _queryMap.values.any((q) => q.stateStream.value?.isFetching);
  }

  Query<T> getQueryByKey<T>(String key) {
    return _queryMap[key];
  }

  Query<T> createQuery<T>(QueryConfig config) {
    var query = getQueryByKey<T>(config.queryKey);
    if (query == null) {
      query = Query<T>(config);
      _queryMap[config.queryKey] = query;
    }
    return query;
  }

  Function watchQuery<T>(QueryConfig config, QueryResultListener<T> listener) {
    final queryObserver = QueryObserver<T>(config, this);
    return queryObserver.subscribe(listener);
  }

  Future<T> fetchQuery<T>(String key, Function queryFn) {
    var query = getQueryByKey<T>(key);

    if (query == null) {
      query = createQuery<T>(
        QueryConfig(
          queryKey: key,
          fn: queryFn,
        ),
      );
    }

    return query.fetch();
  }

  void invalidateQueries(String key) {
    final query = getQueryByKey(key);
    if (query != null) {
      query.invalidate();
      query.fetch();
    }
  }

  void clear() {
    _queryMap.values.forEach((q) => q.dispose());
    _queryMap = {};
  }
}

class QueryCacheProvider extends StatefulWidget {
  final Widget child;
  QueryCacheProvider({this.child});

  @override
  State<StatefulWidget> createState() => QueryCacheProviderState();
}

class QueryCacheProviderState extends State<QueryCacheProvider> {
  QueryCache _queryCache;

  @override
  void initState() {
    _queryCache = QueryCache();
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Provider.value(
      value: _queryCache,
      child: widget.child,
    );
  }
}
