import 'dart:async';

import 'package:flutter_query/src/query.dart';
import 'package:flutter_query/src/query_cache.dart';

typedef void QueryResultListener<T>(QueryResult<T> state);

class QueryObserver<T> {
  QueryConfig config;
  QueryCache cache;
  int staleTime = 0;
  bool isStale = true;

  QueryResultListener<T> _listener;
  Query<T> _query;
  StreamSubscription _staleTimer;

  QueryObserver(
    this.config,
    this.cache,
  ) {
    _query = cache.getQueryByKey(config.queryKey);

    if (_query == null) {
      _query = cache.createQuery(config);
    }
  }

  Function subscribe(QueryResultListener<T> listener) {
    _listener = listener;
    if (_staleTimer == null) {
      _setupStaleTimer();
    }
    final subscription = _query.stateStream.listen(onQueryUpdate);
    if (config.enabled) {
      _query.fetch();
    } else {
      _optionalFetch();
    }

    return () {
      subscription.cancel();
      _listener = null;
      _clearStaleTimer();
    };
  }

  void onQueryUpdate(QueryState<T> state) {
    _notify(state);
  }

  void _notify(QueryState<T> state) {
    if (state.isSuccess && config.onSuccess != null) {
      config.onSuccess(state.data);
    } else if (state.isError && config.onError != null) {
      config.onError(state.data);
    }

    if ((state.isSuccess || state.isError) && config.onSettled != null) {
      config.onSettled(state.data);
    }

    _listener(QueryResult<T>(
      data: state.data,
      err: state.snapshot?.hasError ?? false ? state.snapshot?.error : null,
      failureCount: state.failureCount,
      isError: state.isError,
      isFetching: state.isFetching,
      isIdle: state.isIdle,
      isLoading: state.isLoading,
      isSuccess: state.isSuccess,
      isFetchingMore: state.isFetchingMore,
      isStale: false,
      fetchMore: _query.config.isInfinite
          ? () => _query.fetch()
          : () => Future.value(),
      canFetchMore: state.canFetchMore,
    ));
  }

  void _clearStaleTimer() {
    if (_staleTimer != null) {
      _staleTimer.cancel();
    }
  }

  void _setupStaleTimer() {
    _staleTimer = Future.delayed(
      Duration(milliseconds: config.staleTime),
    ).asStream().listen((_) {
      isStale = true;
      _notify(_query.stateStream.value);
    });
  }

  void _optionalFetch() {
    if (config.enabled && isStale) {
      _query.fetch();
    }
  }
}

class QueryResult<T> {
  T data;
  dynamic err;
  int failureCount;
  bool isError;
  bool isFetchedAfterInit;
  bool isFetching;
  bool isFetchingMore;
  bool isIdle;
  bool isLoading;
  bool isPreviousData;
  bool isStale;
  bool isSuccess;
  bool canFetchMore;
  Function fetchMore;
  Function refetch;
  Function remove;

  QueryResult({
    this.data,
    this.err,
    this.failureCount,
    this.isError,
    this.isFetchedAfterInit,
    this.isFetching,
    this.isFetchingMore,
    this.isIdle,
    this.isLoading,
    this.isPreviousData,
    this.isStale,
    this.isSuccess,
    this.refetch,
    this.remove,
    this.canFetchMore,
    this.fetchMore,
  });
}
